#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "tri.h"

/*
  The subroutine rot rotates coordinate frames based on the Euler
  angles alpha, beta and gamma (see Edmonds page 7 and page 53).
  The input theta and phi are the coordinates in the geographic
  coordinate system. The returned values of theta and phi are
  the coordinates in the equatorial coordinate system.
  The returned value of theta is between 0 and pi, and the returned
  value of phi is between 0 and 2*pi.
  The inverse operation, i.e. from the rotated (equatorial)
  frame back to the original (geographic) frame is performed
  by replacing alpha by -gamma, beta by -beta, and gamma by -alpha
  in the call to the subroutine rot.
*/

void rot(float alpha,float beta,float gamma,float *theta,float *phi,float *zeta)
{
  int i,j;
  float r_hat[3],rp_hat[3],k_hat[3],kp_hat[3],rotation[3][3];
  float sint,cost,sinp,cosp,sina,cosa,sinb,cosb,sing,cosg,sinz,cosz;

  sint=sin(*theta); cost=cos(*theta);
  sinp=sin(*phi); cosp=cos(*phi);
  sina=sin(alpha); cosa=cos(alpha);
  sinb=sin(beta); cosb=cos(beta);
  sing=sin(gamma); cosg=cos(gamma);
  sinz=sin(*zeta); cosz=cos(*zeta);
  r_hat[0]=sint*cosp; r_hat[1]=sint*sinp; r_hat[2]=cost;
  k_hat[0]=-(sinz*cost*cosp+cosz*sinp);
  k_hat[1]=cosz*cosp-sinz*cost*sinp;
  k_hat[2]=sinz*sint;
  rotation[0][0]=cosg*cosb*cosa-sing*sina;
  rotation[0][1]=cosg*cosb*sina+sing*cosa;
  rotation[0][2]=-cosg*sinb;
  rotation[1][0]=-sing*cosb*cosa-cosg*sina;
  rotation[1][1]=-sing*cosb*sina+cosg*cosa;
  rotation[1][2]=sing*sinb;
  rotation[2][0]=sinb*cosa;
  rotation[2][1]=sinb*sina;
  rotation[2][2]=cosb;
  for(i=0;i<3;i++) {
    rp_hat[i]=0.0;
    kp_hat[i]=0.0;
    for(j=0;j<3;j++) {
      rp_hat[i]+=rotation[i][j]*r_hat[j];
      kp_hat[i]+=rotation[i][j]*k_hat[j];
    }
  }
  *theta=atan2(sqrt(rp_hat[0]*rp_hat[0]+rp_hat[1]*rp_hat[1]),rp_hat[2]);
  *phi=atan2(rp_hat[1],rp_hat[0]);
  reduce(theta,phi); 
  sint=sin(*theta); cost=cos(*theta);
  sinp=sin(*phi); cosp=cos(*phi);
  sinz=-(kp_hat[0]*cost*cosp+kp_hat[1]*cost*sinp-kp_hat[2]*sint);
  cosz=-kp_hat[0]*sinp+kp_hat[1]*cosp;
  *zeta=atan2(sinz,cosz);
}

void reduce(float *pth,float *pph)
{
/* force theta between 0 and pi, and phi between 0 and 2*pi */
  int i;
  float th,ph;
  
  th=*pth; ph=*pph;
  i=abs((int)(ph/PI2));
  if(ph<0.0)
    ph+=(i+1)*PI2;
  else if(ph>PI2) ph-=i*PI2;
  *pph=ph;
  if(th<0.0 || th>PI) {
    i=(int)(th/PI);
    if(th>0.0) {
      if(ODD(i)) {
        th=(i+1)*PI-th;
        ph=(ph<PI)?ph+PI:ph-PI;
      }
      else th-=i*PI;
    }
    else {
      if(EVEN(i)) {
        th=-th+i*PI;
        ph=(ph<PI)?ph+PI:ph-PI;
      }
      else th-=i*PI;
    }
    *pth=th; *pph=ph;
  }
}

/*
  Given the source colatitude sco, the source longitude slo,
  the receiver colatitude rco, the receiver longitude rlo,
  the subroutine euler calculates the angular distance delta between
  the source and the receiver, in addition to the Euler angles
  alpha, beta and gamma.
  Note:the source can't be at equator but receiver is ok. 8/18/94
*/

void euler(float sco,float slo,float rco,float rlo,float *delta,
float *alpha,float *beta,float *gamma)
{
  float az,sina,cosa,eta,mu,ep;

  *delta=acos(cos(sco)*cos(rco)+sin(sco)*sin(rco)*cos(rlo-slo));
  sina=sin(rlo-slo)*sin(rco)/sin(*delta);
  cosa=(cos(rco)-cos(sco)*cos(*delta))/(sin(sco)*sin(*delta));
  az=atan2(sina,cosa);
  if(fabs(cosa)<1.0e-5) cosa=1.0e-5;
  eta=atan(cos(sco)/(cosa*sin(sco)));
  if(fabs(eta)<1.0e-5)
    eta=1.0e-5;
  else if(eta<0.0) eta+=PI;
  *gamma=PI/2.0+eta;
  mu=atan2(sina*sin(sco),cos(sco)/sin(eta));
  *beta=PI/2.0-mu;
  ep=atan2(sina*sin(eta),cos(eta)/sin(sco));
  *alpha=-PI/2.0+slo-ep;
}

float *vector(int nl,int nh)
{
	float *v;

	v=(float *)malloc((unsigned) (nh-nl+1)*sizeof(float));
	if (!v) nrerror("allocation failure in vector()");
	return v-nl;
}

double *dvector(int nl,int nh)
{
	double *v;

	v=(double *)malloc((unsigned) (nh-nl+1)*sizeof(double));
	if (!v) nrerror("allocation failure in dvector()");
	return v-nl;
}



float **matrix(int nrl,int nrh,int ncl,int nch)
{
	int i;
	float **m;

	m=(float **) malloc((unsigned) (nrh-nrl+1)*sizeof(float*));
	if (!m) nrerror("allocation failure 1 in matrix()");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(float *) malloc((unsigned) (nch-ncl+1)*sizeof(float));
		if (!m[i]) nrerror("allocation failure 2 in matrix()");
		m[i] -= ncl;
	}
	return m;
}

double **dmatrix(int nrl,int nrh,int ncl,int nch)
{
	int i;
	double **m;

	m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
	if (!m) nrerror("allocation failure 1 in dmatrix()");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
		if (!m[i]) nrerror("allocation failure 2 in dmatrix()");
		m[i] -= ncl;
	}
	return m;
}

void free_vector(float *v,int nl,int nh)
{
	free((char*) (v+nl));
}

void free_dvector(double *v,int nl,int nh)
{
	free((char*) (v+nl));
}



void free_matrix(float **m,int nrl,int nrh,int ncl,int nch)
{
	int i;

	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

void free_dmatrix(double **m,int nrl,int nrh,int ncl,int nch)
{
	int i;

	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

void nrerror(char *error_text)
{

        fprintf(stderr,"SPH Triangular Tesselation run-time error...\n");
        fprintf(stderr,"%s\n",error_text);
        fprintf(stderr,"...now exiting to system...\n");
        exit(1);
}
